package ap.parser;

import ap.terfor.{ConstantTerm, Conjunction, Quantifier, OneTerm}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.util.{Debug, Logic, PlainRange}
import ap.proof.goal.Goal
import ap.basetypes.IdealInt
import ApInput._
import ApInput.Absyn._

object PlainTranslator {

  private val AC = Debug.AC_PARSER
  
  class ParseException(msg : String) extends Exception(msg)
  class TranslationException(msg : String) extends Exception(msg)
  
  def apply(input : java.io.Reader) : Goal = {
    val l = new Yylex(input)
    val p = new parser(l)
    
    val api = try { p.pAPI } catch {
      case e : Exception =>
        throw new ParseException("At line " + String.valueOf(l.line_num()) +
                                 ", near \"" + l.buff() + "\" :" +
                                 "     " + e.getMessage())
    }
    
    apply(api)
  }
  
  def apply(api : API) : Goal = {
    val t = new PlainTranslator
    t collectDeclarations api
    t translateProblem api
  }
  
}


class PlainTranslator {
  
  private val env = new Environment
  
  /** Implicit conversion so that we can get a scala-like iterator from a
    * a Java list */
  private implicit def toWrappedList[A](list : java.util.LinkedList[A]) =
    new scala.collection.jcl.LinkedList(list)
  
  //////////////////////////////////////////////////////////////////////////////
    
  def collectDeclarations(api : API) : unit = api match {
    case api : BlockList =>
      for (block <- api.listblock_.elements) block match {
        case block : FunctionDecls =>
          for (decl <- block.listdeclfunc_.elements)
            collectDeclFunC(decl,
                            (id) => env.addConstant(new ConstantTerm(id),
                                                    Environment.Universal))
        case block : ExConstants =>
          for (decl <- block.listdeclexconstantc_.elements)
            collectDeclExConstantC(decl,
                                   (id) => env.addConstant(new ConstantTerm(id),
                                                           Environment.Existential))
        case _ : Problem => /* nothing */
      }
  }

  //////////////////////////////////////////////////////////////////////////////

  private def collectDeclFunC(decl : DeclFunC, addCmd : String => unit) : unit =
    collectDeclarations(decl.asInstanceOf[DeclFun].declvarconstc_, addCmd)
    
  private def collectDeclExConstantC(decl : DeclExConstantC, addCmd : String => unit) : unit =
    collectDeclarations(decl.asInstanceOf[DeclExConstant].declvarconstc_, addCmd)

  private def collectDeclBinder(decl : DeclBinder,
                                addCmd : String => unit) : unit = decl match {
    case decl : DeclBinder1 => collectDeclarations(decl.declvarconstc_, addCmd)
    case decl : DeclBinderM => for (decl <- decl.listdeclvarconstc_.elements) 
                                 collectDeclarations(decl, addCmd)
  }

  private def collectDeclarations(decl : DeclVarConstC,
                                  addCmd : String => unit) : unit = decl match {
    case decl : DeclVarConst => { 
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(PlainTranslator.AC, decl.type_.isInstanceOf[TypeInt])
      //////////////////////////////////////////////////////////////////////////
      for (id <- decl.listident_.elements) addCmd(id)
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  
  def translateProblem(api : API) : Goal = api match {
    case api : BlockList => {
      api.listblock_.filter(_.isInstanceOf[Problem]) match {
        case Seq(problem : Problem) => {
          val f = translateFormula(problem.formula_)
          Goal(f, env.universalConstants, env.order)
        }
        case _ => throw new PlainTranslator.TranslationException(
                             "Found zero or more than one \\problem blocks")
      }
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  def translateFormula(f : Formula) : ap.terfor.Formula = f match {
    case f : ForEqv =>
      translateBinConnective(f.formula_1, f.formula_2,
                             Conjunction.eqv(_, _, env.order))
    case f : ForImp =>
      translateBinConnective(f.formula_1, f.formula_2,
                             Conjunction.implies(_, _, env.order))
    case f : ForOr =>
      translateBinConnective(f.formula_1, f.formula_2,
                             (f1,f2) => Conjunction.disj(Array(f1, f2), env.order))
    case f : ForAnd =>
      translateBinConnective(f.formula_1, f.formula_2,
                             (f1,f2) => Conjunction.conj(Array(f1, f2), env.order))
    case f : ForNot =>
      translateUnConnective(f.formula_, Conjunction.negate(_, env.order))
    case f : ForQuant =>
      translateQuant(f)
    case _ : ForTrue => Conjunction.TRUE
    case _ : ForFalse => Conjunction.FALSE
    case f : ForRel => {
      val diff = sum(translateTerm(f.term_1), translateTerm(f.term_2), IdealInt.MINUS_ONE)
      f.relsym_ match {
        case _ : RelEq => EquationConj(Array(diff), env.order)
        case _ : RelNEq => NegEquationConj(Array(diff), env.order)
      }
    }
  }
  
  private def translateUnConnective(f : Formula,
                                    con : (ap.terfor.Formula) => ap.terfor.Formula)
              : ap.terfor.Formula =
    con(translateFormula(f))
  
  private def translateBinConnective(f1 : Formula, f2 : Formula,
                                     con : (ap.terfor.Formula, ap.terfor.Formula) =>
                                                             ap.terfor.Formula)
              : ap.terfor.Formula =
    con(translateFormula(f1), translateFormula(f2))
  
  private def translateQuant(f : ForQuant) : ap.terfor.Formula = {
    val quant = f.quant_ match {
      case _ : QuantAll => Quantifier.ALL
      case _ : QuantEx => Quantifier.EX
    }
    
    // add the bound variables to the environment and record their number
    var quantNum : Int = 0
    collectDeclBinder(f.declbinder_,
                      (id) => { quantNum = quantNum + 1; env pushVar id })
    
    val res = translateUnConnective(f.formula_,
                                    Conjunction.quantify(Array(quant), _, env.order))

    // pop the variables from the environment
    for (_ <- PlainRange(quantNum)) env.popVar
    
    res
  }

  //////////////////////////////////////////////////////////////////////////////

  private def translateTerm(t : Term) : LinearCombination = t match {
    case t : TermPlus =>
      translateBinOp(t.term_1, t.term_2, sum(_, _, IdealInt.ONE))
    case t : TermMinus =>
      translateBinOp(t.term_1, t.term_2, sum(_, _, IdealInt.MINUS_ONE))
    case t : TermMult =>
      translateBinOp(t.term_1, t.term_2, mult _)
    case t : TermUnPlus =>
      translateUnOp(t.term_, (lc) => lc)
    case t : TermUnMinus =>
      translateUnOp(t.term_, _ * IdealInt.MINUS_ONE)
    case t : TermId => (env lookupSym t.ident_) match {
      case Environment.Constant(c, _) => LinearCombination(c, env.order)
      case Environment.Variable(v) => LinearCombination(v, env.order)
    }
    case t : TermLit =>
      LinearCombination(IdealInt(t.intlit_))
  }

  private def sum(lc1 : LinearCombination, lc2 : LinearCombination, coeff : IdealInt)
                                                     : LinearCombination =
    LinearCombination.sum(Array((IdealInt.ONE, lc1), (coeff, lc2)), env.order)
                                                       
  private def mult(lc1 : LinearCombination, lc2 : LinearCombination)
              : LinearCombination = (isLiteral(lc1), isLiteral(lc2)) match {
    case (true, _) => lc2 * (lc1 get OneTerm)
    case (_, true) => lc1 * (lc2 get OneTerm)
    case _ => throw new PlainTranslator.TranslationException(
                        "Do not know how to multiply " + lc1 + " with " + lc2)
  }
  
  private def isLiteral(lc : LinearCombination) : boolean = lc.size match {
    case 0 => true
    case 1 => (lc(0) _2) == OneTerm
    case _ => {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertPost(PlainTranslator.AC,
                       !Logic.forall(for ((_, t) <- lc.elements) yield (t == OneTerm)))
      //////////////////////////////////////////////////////////////////////////
      false
    }
  }

  private def translateUnOp(t : Term,
                            op : (LinearCombination) => LinearCombination)
                                                          : LinearCombination =
    op(translateTerm(t))

  private def translateBinOp(t1 : Term, t2 : Term,
                             op : (LinearCombination, LinearCombination)
                                    => LinearCombination) : LinearCombination =
    op(translateTerm(t1), translateTerm(t2))

}